/*
* dvb-aspect.c:  Tool to grab the aspect ratio of video played on a dvb ff (av7110)
*
* Copyright (C) 2006 Jens Vogel (jens.vogel@justmail.de)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
* Or, point your browser to http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>
#include <float.h>
#include <math.h>
#include <memory.h>
#include <getopt.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/dvb/video.h>
#include <linux/videodev.h>
#include <linux/dvb/frontend.h>

#define DEFAULT_VIDEO_DEVICE "/dev/video0"
#define DEFAULT_DVB_DEVICE  "/dev/dvb/adapter0"
#define DEFAULT_MAX_LOGO_WIDTH 0.2
#define DEFAULT_PAN_TONE_TOLERANCE 8
#define DEFAULT_OVERSCAN 0.03
#define DEFAULT_POLL_RATE 0.08
#define DEFAULT_DELAY 25

typedef struct 
{
  float       fMinAspect;
  const char* pszName;
} SAspectInfo;
SAspectInfo aAspectInfo[] =
{
  { 1.46,     "4:3" },
  { 1.66,     "L14:9" },
  { 1.90,     "L16:9" },
  { FLT_MAX,  "L>16:9" }
};
int naAspectInfo = sizeof(aAspectInfo) / sizeof(SAspectInfo);

double GetTime()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}

int main(int argc, char **argv)
{
  char *pszDevVideo = DEFAULT_VIDEO_DEVICE;
  char *pszDvb = DEFAULT_DVB_DEVICE;
  unsigned char iPanToneBlackMax = 0;
  int iPanToneTolerance = DEFAULT_PAN_TONE_TOLERANCE;
  int bNativeAspectOutput = 0;
  float fMaxLogoWidth = DEFAULT_MAX_LOGO_WIDTH;
  float fOverscan = DEFAULT_OVERSCAN;
  float fPollRate = DEFAULT_POLL_RATE;
  unsigned nDelay = DEFAULT_DELAY;
  int bFrontendHasLockTest = 0;
  int bOverzoom = 0;

  int bMoreOptions = true;
  opterr = 0;
  do
  {
    static struct option aOptions[] =
    {
      { "dev_video", required_argument, 0, 'v' },
      { "dev_dvb", required_argument, 0, 'd' },
      { "black", required_argument, 0, 'b' },
      { "pan_tone_tolerance", required_argument, 0, 't' },
      { "logo_width", required_argument, 0, 'l' },
      { "overscan", required_argument, 0, 's' },
      { "poll_rate", required_argument, 0, 'p' },
      { "delay", required_argument, 0, 'e' },
      { "aspect", no_argument, 0, 'a' },
      { "help", no_argument, 0, 'h' },
      { NULL }
    };
    switch(getopt_long(argc, argv, "v:d:b:t:l:s:p:e:afoh", aOptions, NULL))
    {
    case 'v':
      pszDevVideo = optarg;
      break;
    case 'd':
      pszDvb = optarg;
      break;
    case 'b':
      sscanf(optarg, "%x", &iPanToneBlackMax);
      break;
    case 't':
      sscanf(optarg, "%x", &iPanToneTolerance);
      break;
    case 'l':
      sscanf(optarg, "%g", &fMaxLogoWidth);
      fMaxLogoWidth /= 100;
      break;
    case 's':
      sscanf(optarg, "%g", &fOverscan);
      fOverscan /= 100;
      break;
    case 'p':
      sscanf(optarg, "%g", &fPollRate);
      break;
    case 'e':
      sscanf(optarg, "%u", &nDelay);
      break;
    case 'a':
      bNativeAspectOutput = 1;
      break;
    case 'f':
      bFrontendHasLockTest = 1;
      break;
    case 'o':
      bOverzoom = 1;
      break;
    case 'h':
      printf("Usage: %s [<Options>]\n"
             "  -v, --dev_video=<Device>  Video device\n"
             "                            (default is " DEFAULT_VIDEO_DEVICE ")\n" 
             "  -d, --dev_dvb=<Device>    DVB device\n"
             "                            (default is " DEFAULT_DVB_DEVICE ")\n"
             "  -t, --pan_tone_tolerance  Maximum difference of a pixel from the most gray tone pixels within\n"
             "   =<Value>                 a pan vector (hex; grayscale; 0..ff; default is %02x).\n"
             "                            This option and its value is ignored if -b or --black is given\n"
             "  -b, --black=<Value>       Enable pan vector detection by a given lowest value of a pixel,\n"
             "                            interpreted as black (hex; grayscale; 0..ff; choose 20 or so).\n"
             "                            Otherwise the gray tone of the pan vector is detected automaticly\n"
             "                            (may be white, green or whatever just as well).\n"
             "  -s, --overscan=<Value>    The width of a frame around the picture which is ignored during\n"
             "                            analyzing of the pan vectors (real number; percent; 0..100; default is %g)\n"
             "  -l, --logo_width=<Value>  Maximum width of a range with wrong pixels (a logo) within a pan\n"
             "                            vector (real number; percent; 0..100; default is %g)\n"
             "  -p, --poll_rate=<Value>   Poll rate (frequency of picture analyzing; real number; seconds; >0;\n"
             "                            (default is %g; lower values than 0.08 makes no sense at the moment)\n"
             "  -e, --delay=<Value>       Delay output until last <Value> analyzed pictures produces the\n"
             "                            same result (multiply with poll_rate to get delay time;\n"
             "                            decimal number; >= 0; default is %u)\n"
             "  -a, --aspect              Output aspect ratio in addition as number\n"
             "                            (e.g \"133 4:3\")\n"
             "  -f                        Enable 'frontend has lock' test\n"
             "                            (no longer useful, because of black screen detection, but...)\n"
             "  -o                        Enable L>16:9 output (normaly this makes no sense)\n"
             "  -h, --help                Displays this page (and exits)\n",
             argv[0],
             DEFAULT_PAN_TONE_TOLERANCE,
             DEFAULT_OVERSCAN*100,
             DEFAULT_MAX_LOGO_WIDTH*100,
             DEFAULT_POLL_RATE,
             DEFAULT_DELAY
             );
      return 0;
      break;
    case '?':
      fprintf(stderr, "%s: error - unkown option %s [arg #%d] - ignored\n", argv[0], argv[optind-1], optind-1);
      break;
    case ':':
      fprintf(stderr, "%s: error - missing argument for option %s [arg #%d] - ignored\n", argv[0], argv[optind-1], optind-1);
      break;
    default:
      bMoreOptions = false;
      break;
    }
  }
  while(bMoreOptions);
  if (optind < argc)
    fprintf(stderr, "%s: error - extra arguments '%s'... [arg #%d...] - ignored\n", argv[0], argv[optind], optind);

  if (!bOverzoom)
  {
      naAspectInfo--;
      aAspectInfo[naAspectInfo-1].fMinAspect = aAspectInfo[naAspectInfo].fMinAspect;
  }

  char pszDvbVideo[100];
  strcat(strcpy(pszDvbVideo, pszDvb), "/video0");
  char pszDvbFrontend[100];
  if (bFrontendHasLockTest)
    strcat(strcpy(pszDvbFrontend, pszDvb), "/frontend0");

  int devDvbVideo = open(pszDvbVideo,O_RDONLY|O_NONBLOCK);
  if (devDvbVideo < 0)
    fprintf(stderr, "Error: Can't open dvb video device '%s' (%s)\n", pszDvbVideo, strerror(errno));
  else
  {
    int devDvbFrontend;
    if (bFrontendHasLockTest)
      devDvbFrontend = open(pszDvbFrontend,O_RDONLY|O_NONBLOCK);
    if (bFrontendHasLockTest && (devDvbFrontend < 0))
      fprintf(stderr, "Error: Can't open dvb frontend device '%s' (%s)\n", pszDvbFrontend, strerror(errno));
    else
    {
      int devVideo = open(pszDevVideo, O_RDWR);
      if (devVideo < 0)
        fprintf(stderr, "Error: Can't open video device '%s' (%s)\n", pszDevVideo, strerror(errno));
      else 
      {
        struct video_mbuf mbufVideo;
        if (ioctl(devVideo, VIDIOCGMBUF, &mbufVideo) != 0) 
          fprintf(stderr, "Error: Can't open video buffer on video device '%s' (%s)\n", pszDevVideo, strerror(errno));
        else
        {
          unsigned char *pbVideo = (unsigned char *)mmap(0, mbufVideo.size, PROT_READ | PROT_WRITE, MAP_SHARED, devVideo, 0);
          if (!pbVideo || (pbVideo == (unsigned char *)-1))
            fprintf(stderr, "Error: Can't map video buffer of video device '%s' (%s)\n", pszDevVideo, strerror(errno));
          else
          {
            struct video_capability capabilityVideo;
            if (ioctl(devVideo, VIDIOCGCAP, &capabilityVideo))
              fprintf(stderr, "Error: Can't query caps of video device '%s' (%s)\n", pszDevVideo, strerror(errno));
            else
            {
              struct video_mmap mmapVideo;
              mmapVideo.frame = 0;
              mmapVideo.width = capabilityVideo.maxwidth;
              mmapVideo.height = capabilityVideo.maxheight;
              mmapVideo.format = VIDEO_PALETTE_GREY;
              int sizeVideo = mmapVideo.width * mmapVideo.height;
              int bAnamorph = 0;
              int iAspectInfo = -1;
              float fNativeAspectOutput = 0;
              char** ppszDelay = (char**)calloc(nDelay, sizeof(char*));
              for (int i = 0; i < nDelay; i++)
                ppszDelay[i] = strdup("");
              int bDelay = 1;
              int bError = 0;
              while (!bError)
              {
                int bFormatChanged = 0;
                fe_status_t FrontendStatus;
                double fStartTime = GetTime();
                if (bFrontendHasLockTest && (bError = (ioctl(devDvbFrontend, FE_READ_STATUS, &FrontendStatus) < 0)))
                  fprintf(stderr, "Error: Can't get status of device '%s' (%s)\n", pszDvbFrontend, strerror(errno));
                else
                {
                  if (!bFrontendHasLockTest || (FrontendStatus & FE_HAS_LOCK))
                  {
                    struct video_status statusVideo;
                    if (bError = (ioctl(devDvbVideo, VIDEO_GET_STATUS, &statusVideo) < 0)) 
                      fprintf(stderr, "Error: Can't get status of device '%s' (%s)\n", pszDvbVideo, strerror(errno));
                    else
                    {
                      switch (statusVideo.video_format)
                      {
                      case VIDEO_FORMAT_16_9:
                      case VIDEO_FORMAT_221_1:
                        if (!bAnamorph)
                        {
                          bFormatChanged = 1;
                          bAnamorph = 1;
                        }
                        break;
                      case  VIDEO_FORMAT_4_3:
                      default:
                        if (bAnamorph)
                        {
                          iAspectInfo = -1;
                          bAnamorph = 0;
                        }
                        if (bError = (  ioctl(devVideo, VIDIOCMCAPTURE, &mmapVideo)
                          | ioctl(devVideo, VIDIOCSYNC, &mmapVideo.frame)))
                          fprintf(stderr, "Error: Can't grab on video device '%s' (%s)\n", pszDevVideo, strerror(errno));
                        else
                        {
                          const int nMaxLogoWidth = int(mmapVideo.width*fMaxLogoWidth);
                          const int nYoverscan = int (mmapVideo.height*fOverscan);
                          const int nXoverscan = int (mmapVideo.width*fOverscan);
                          
                          unsigned char iPanToneMin;
                          unsigned char iPanToneMax;
                          
                          int nActiveLines = 0;
                          int iYtop = nYoverscan;
                          
                          if (iPanToneBlackMax)
                          {
                            iPanToneMin = 0;
                            iPanToneMax = iPanToneBlackMax;
                          }
                          else
                          {
                            int aHistogram[256];
                            memset(aHistogram, 0, sizeof(aHistogram));
                            unsigned char *pbVideoLine = pbVideo+iYtop*mmapVideo.width;
                            for (int iX = nXoverscan; iX < mmapVideo.width-nXoverscan; iX++)
                              aHistogram[pbVideoLine[iX]]++;
                            int iXmax = 0;
                            for (int i = 0; i < 256; i++)
                              if (aHistogram[i] > aHistogram[iXmax])
                                iXmax = i;
                              if (iXmax - iPanToneTolerance < 0)
                                iPanToneMin = 0;
                              else
                                iPanToneMin = iXmax - iPanToneTolerance;
                              if (iXmax + iPanToneTolerance > 255)
                                iPanToneMax = 255;
                              else
                                iPanToneMax = iXmax + iPanToneTolerance;
                          }
                          
                          while (iYtop < mmapVideo.height/2)
                          {
                            unsigned char *pbVideoLine = pbVideo+iYtop*mmapVideo.width;
                            int iXleft = nXoverscan;
                            while (    (iXleft < (mmapVideo.width-nXoverscan)) 
                              && (pbVideoLine[iXleft] >= iPanToneMin)
                              && (pbVideoLine[iXleft] <= iPanToneMax)  )
                              iXleft++;
                            int iXright = mmapVideo.width-nXoverscan-1;
                            while (    (iXright > iXleft) 
                              && (pbVideoLine[iXright] >= iPanToneMin)
                              && (pbVideoLine[iXright] <= iPanToneMax) )
                              iXright--;
                            if (iXright - iXleft > nMaxLogoWidth)
                              break;
                            iYtop++;
                          }
                          if (iYtop < mmapVideo.height/2)
                          {
                            int iYbottom = nYoverscan;
                            while (iYbottom < mmapVideo.height/2)
                            {
                              unsigned char *pbVideoLine = pbVideo+(mmapVideo.height-1-iYbottom)*mmapVideo.width;
                              int iXleft = nXoverscan;
                              while (    (iXleft < (mmapVideo.width-nXoverscan)) 
                                && (pbVideoLine[iXleft] >= iPanToneMin)
                                && (pbVideoLine[iXleft] <= iPanToneMax)  )
                                iXleft++;
                              int iXright = mmapVideo.width-nXoverscan-1;
                              while (    (iXright > iXleft) 
                                && (pbVideoLine[iXright] >= iPanToneMin)
                                && (pbVideoLine[iXright] <= iPanToneMax) )
                                iXright--;
                              if (iXright - iXleft > nMaxLogoWidth)
                                break;
                              iYbottom++;
                            }
                            if (iYbottom < mmapVideo.height/2)
                            {
                              int iBlackLines;
                              if (iYtop > iYbottom)
                                iBlackLines = 2*iYbottom;
                              else
                                iBlackLines = 2*iYtop;
                              nActiveLines = mmapVideo.height - iBlackLines;
                            }
                          }
                          
                          if (nActiveLines > 0)
                          {
                            const float fLinearWidth = 4./3.*mmapVideo.height;
                            if (iAspectInfo < 0)
                              bFormatChanged = 1;
                            else
                            {
                              const int nAspectHyst = 2;
                              int nMinActiveLines = (int)(fLinearWidth/aAspectInfo[iAspectInfo].fMinAspect) - nAspectHyst;
                              int nMaxActiveLines =   (iAspectInfo > 0)
                                ? (int)(fLinearWidth/aAspectInfo[iAspectInfo-1].fMinAspect) + nAspectHyst
                                : mmapVideo.height;
                              bFormatChanged = (nActiveLines < nMinActiveLines) || (nActiveLines > nMaxActiveLines); 
                            }
                            float fAspect = fLinearWidth/nActiveLines;
                            if (bNativeAspectOutput)
                            {
                              float fNewNativeAspectOutput = floor(fAspect*100); 
                              if (!bFormatChanged)
                                bFormatChanged = fNewNativeAspectOutput != fNativeAspectOutput;
                              fNativeAspectOutput = fNewNativeAspectOutput;
                            }
                            if (bFormatChanged)
                            {
                              iAspectInfo = 0;
                              while (fAspect > aAspectInfo[iAspectInfo].fMinAspect)
                                iAspectInfo++;
                            }
                          }
                        }
                        break;
                      }
                    }
                  }
                }
                if (!bError)
                {
                  if (bAnamorph)
                  {
                    if (bFormatChanged)
                    {
                      printf("16:9\n");
                      fflush(stdout);
                    }
                    free(ppszDelay[0]);
                    ppszDelay[0] = strdup("");
                    bDelay = 1;
                  }
                  else if (iAspectInfo >= 0)
                  {
                    static char szOutput[80];
                    if (bFormatChanged)
                    {
                      const char* pszAspect = aAspectInfo[iAspectInfo].pszName;
                      if (bNativeAspectOutput)
                        sprintf(szOutput, "%.0f %s", fNativeAspectOutput, pszAspect);
                      else
                        sprintf(szOutput, "%s", pszAspect);
                    }
                    int iDelay = 0;
                    if (!nDelay || !strcmp(ppszDelay[0], ""))
                    {
                      iDelay = nDelay;
                      bDelay = 1;
                      for (int i = 0; i < nDelay; i++)
                      {
                        free(ppszDelay[i]);
                        ppszDelay[i] = strdup(szOutput);
                      }
                    }
                    else
                      while ((iDelay < nDelay) && !strcmp(ppszDelay[iDelay], szOutput))
                        iDelay++;
                      if (iDelay < nDelay)
                        bDelay = 1;
                      else if (bDelay)
                      {
                        bDelay = 0;
                        printf("%s\n", szOutput);
                        fflush(stdout);
                      }
                      if (nDelay)
                      {
                        free(ppszDelay[nDelay-1]);
                        memmove(ppszDelay+1, ppszDelay, sizeof(char*)*(nDelay-1));
                        *ppszDelay = strdup(szOutput);
                      }
                  }
                  
                  double fDelay = fPollRate - (GetTime() - fStartTime) - 0.001;
                  if (fDelay > 0)
                    usleep(unsigned(fDelay*1000000));
                }
              }
              munmap(pbVideo, mbufVideo.size);
            }
          }
        }
        close(devVideo);
      }
      if (bFrontendHasLockTest)
        close(devDvbFrontend);
    }
    close(devDvbVideo);
  }
  return 0;
}
